from django.apps import AppConfig as BaseAppConfig


class AppConfig(BaseAppConfig):
    name = "frontend_kit"
    verbose_name = "Django Frontend Kit"
    default_auto_field = "django.db.models.BigAutoField"



import json
from abc import ABC, abstractmethod
from functools import cache
from pathlib import Path
from typing import Any, Generator, Hashable, NamedTuple, Self, cast

from django.conf import settings
from django.templatetags.static import static


class ManifestEntry(NamedTuple):
    name: str
    file: str
    src: str = ""
    is_entry: bool = False
    is_dynamic_entry: bool = False
    import_list: list[str] = []
    asset_list: list[str] = []
    css_list: list[str] = []


class AssetNotFoundError(Exception): ...


class AssetTag(ABC, Hashable):
    src: str

    def __init__(self, src: str) -> None:
        self.src: str = src

    @abstractmethod
    def render(self, nonce: str | None = None) -> str:
        raise NotImplementedError

    def __hash__(self) -> int:
        return hash(self.src)

    def __eq__(self, value: object) -> bool:
        return self.src == cast(Self, value).src

    def __str__(self) -> str:
        return self.src

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(src={self.src!r})"


class ModulePreloadTag(AssetTag):
    def render(self, nonce: str | None = None) -> str:
        return f'<link rel="modulepreload" href="{self.src}" />'


class ModuleTag(AssetTag):
    def render(self, nonce: str | None = None) -> str:
        nonce_attr = f' nonce="{nonce}"' if nonce else ""
        return (
            f'<script type="module" src="{self.src}"{nonce_attr}></script>'
        )


class StyleSheetTag(AssetTag):
    def render(self, nonce: str | None = None) -> str:
        nonce_attr = f' nonce="{nonce}"' if nonce else ""
        return f'<link rel="stylesheet" href="{self.src}"{nonce_attr}>'


class AssetResolver(ABC):
    @abstractmethod
    def get_imports(self, file: str) -> Generator[AssetTag, None, None]: ...


class ViteDevServerAssetResolver(AssetResolver):
    def get_imports(self, file: str) -> Generator[AssetTag, None, None]:
        vite_dev_server_url = getattr(
            settings, "VITE_DEV_SERVER_URL", "http://localhost:5173/"
        )
        static_url = vite_dev_server_url + file
        yield ModuleTag(src=static_url)


class ManifestAssetResolver(AssetResolver):
    def __init__(self, entries: dict[str, ManifestEntry]) -> None:
        self.entries = entries

    def get_imports(self, file: str) -> Generator[AssetTag, None, None]:
        if file not in self.entries:
            raise FileNotFoundError(
                f"File {file} does not exist in manifest, "
                "did you build your Vite project?"
            )
        entry = self.entries[file]
        for js_file in entry.import_list:
            yield ModulePreloadTag(src=static(self.entries[js_file].file))
        yield from self.__get_stylesheets(entry=entry)
        yield ModuleTag(src=static(entry.file))

    def __get_stylesheets(
        self, entry: ManifestEntry
    ) -> Generator[StyleSheetTag, None, None]:
        for css_file in entry.css_list:
            yield StyleSheetTag(src=static(css_file))

        for imported_entry in entry.import_list:
            yield from self.__get_stylesheets(
                entry=self.entries[imported_entry]
            )


class ViteAssetResolver:
    @staticmethod
    def get_imports(file: str) -> Generator[AssetTag, None, None]:
        resolver: AssetResolver
        if hasattr(settings, "DJFK_DEV_ENV") and getattr(
            settings, "DJFK_DEV_ENV", True
        ):
            resolver = ViteDevServerAssetResolver()
        else:
            resolver = ManifestAssetResolver(get_vite_manifest())
        yield from resolver.get_imports(file=file)


@cache
def get_vite_manifest() -> dict[str, ManifestEntry]:
    entries: dict[str, ManifestEntry] = {}
    if not hasattr(settings, "VITE_OUTPUT_DIR"):
        raise RuntimeError(
            "VITE_OUTPUT_DIR is not set in settings.py, please set it to the "
            "output directory of your Vite build"
        )

    vite_output_dir = Path(settings.VITE_OUTPUT_DIR)
    if not vite_output_dir.exists():
        raise RuntimeError(
            f"{settings.VITE_OUTPUT_DIR} does not exist, "
            "please check VITE_OUTPUT_DIR settings to ensure you "
            "pass correct dir"
        )

    manifest_path = vite_output_dir / ".vite" / "manifest.json"
    manifest_content = manifest_path.read_text()
    manifest: dict[str, Any] = json.loads(manifest_content)
    for file, entry in manifest.items():
        entries[file] = ManifestEntry(
            name=entry["name"],
            file=entry["file"],
            src=entry.get("src", ""),
            is_entry=entry.get("isEntry", False),
            is_dynamic_entry=entry.get("isDynamicEntry", False),
            import_list=entry.get("imports", []),
            asset_list=entry.get("assets", []),
            css_list=entry.get("css", []),
        )

    return entries



from __future__ import annotations

import sys
from pathlib import Path
from typing import Any, Generator, cast

from django.conf import settings
from django.http import HttpRequest, HttpResponse
from django.template import RequestContext, loader
from render_block import render_block_to_string

from frontend_kit import utils
from frontend_kit.manifest import (
    AssetTag,
    ModulePreloadTag,
    ModuleTag,
    StyleSheetTag,
    ViteAssetResolver,
)


class FileNotLoadedFromViteError(Exception):
    def __init__(self, file_name: str) -> None:
        super().__init__(f"{file_name} was not included in Vite manifest")


class PageMeta(type):
    def __init__(  # noqa: C901
        cls,
        name: str,
        bases: tuple[type, ...],
        namespace: dict[str, Any],
    ) -> None:
        super().__init__(name, bases, namespace)

        if cls.__name__ == "Page":
            return  # Skip base

        cls._assets: dict[str, list[AssetTag]] = {
            "stylesheets": [],
            "preloads": [],
            "head": [],
            "body": [],
        }

        # Load files from Meta
        head_files = [
            "entry.head.js",
            "entry.head.ts",
        ]
        body_files = [
            "entry.js",
            "entry.ts",
        ]

        if settings.DEBUG and hasattr(settings, "VITE_DEV_SERVER_URL"):
            cls._assets["head"].append(
                ModuleTag(src=f"{settings.VITE_DEV_SERVER_URL}@vite/client")
            )

        seen: set[AssetTag] = set()
        for files, section in ((head_files, "head"), (body_files, "body")):
            for file_name in files:
                try:
                    tags_generator = cls.resolve_import(file_name=file_name)  # type: ignore
                except FileNotFoundError:
                    # ignore if any file is not found
                    continue
                for tag in tags_generator:
                    if tag in seen:
                        continue
                    if isinstance(tag, StyleSheetTag):
                        cls._assets["stylesheets"].append(tag)
                    elif isinstance(tag, ModulePreloadTag):
                        cls._assets["preloads"].append(tag)
                    elif isinstance(tag, ModuleTag):
                        cls._assets[section].append(tag)
                    seen.add(tag)


class Page(metaclass=PageMeta):
    _assets: dict[str, list[AssetTag]] = {}

    def __init__(self) -> None:
        self.stylesheets: list[StyleSheetTag] = []
        self.preload_imports: list[ModulePreloadTag] = []
        self.head_imports: list[ModuleTag] = []
        self.body_imports: list[ModuleTag] = []
        self._collect_inherited_assets()

    def _collect_inherited_assets(self) -> None:
        collected: dict[str, list[AssetTag]] = {
            "stylesheets": [],
            "preloads": [],
            "head": [],
            "body": [],
        }
        seen: set[AssetTag] = set()

        for cls in self.__class__.__mro__:
            if not hasattr(cls, "_assets"):
                continue
            for key, values in cls._assets.items():
                for tag in values:
                    if tag not in seen:
                        collected[key].append(tag)
                        seen.add(tag)

        self.stylesheets = collected["stylesheets"]  # type: ignore
        self.preload_imports = collected["preloads"]  # type: ignore
        self.head_imports = collected["head"]  # type: ignore
        self.body_imports = collected["body"]  # type: ignore

    @classmethod
    def get_template_name(cls) -> str:
        return cls.get_relative_template_name("index.html")

    @classmethod
    def get_relative_template_name(cls, name: str) -> str:
        return str(cls._get_base_path() / name)

    def get_context(self) -> dict[str, Any]:
        return {"page": self}

    def render(
        self,
        *,
        request: HttpRequest,
        relative_template_name: str = "",
    ) -> str:
        template_name = (
            self.get_template_name()
            if not relative_template_name
            else self.get_relative_template_name(name=relative_template_name)
        )
        return str(
            loader.get_template(template_name=template_name).render(
                context=self.get_context(),
                request=request,
            )
        )

    def render_block(
        self,
        *,
        block_name: str,
        request: HttpRequest,
        relative_template_name: str = "",
    ) -> str:
        template_name = (
            self.get_template_name()
            if not relative_template_name
            else self.get_relative_template_name(name=relative_template_name)
        )
        context = RequestContext(request, self.get_context())
        return cast(
            str,
            render_block_to_string(
                template_name=template_name,
                block_name=block_name,
                context=context,
                request=request,
            ),
        )

    def as_response(
        self,
        *,
        request: HttpRequest,
        block_name: str = "",
    ) -> HttpResponse:
        if block_name:
            html = self.render_block(request=request, block_name=block_name)
        else:
            html = self.render(request=request)
        return HttpResponse(content=html.encode())

    @classmethod
    def resolve_import(
        cls, *, file_name: str
    ) -> Generator[AssetTag, None, None]:
        base = cls._get_base_path()
        path = base / file_name if isinstance(file_name, str) else file_name
        if not path.exists():
            raise FileNotFoundError(f"file {file_name} not found")
        if name := cls._get_js_manifest_name(path):
            return ViteAssetResolver.get_imports(file=name)
        raise FileNotLoadedFromViteError(file_name=file_name)

    @classmethod
    def _get_base_path(cls) -> Path:
        return Path(cls._get_file_path()).parent

    @classmethod
    def _get_file_path(cls) -> str:
        mod = sys.modules[cls.__module__]
        path = getattr(mod, "__file__", None)
        if not path:
            raise RuntimeError(f"Can't determine file path for {cls}")
        return str(path)

    @classmethod
    def _get_js_manifest_name(cls, file_path: Path) -> str | None:
        frontend_dir = utils.get_frontend_dir_from_settings()
        if not file_path.exists():
            return None
        return str(file_path.relative_to(Path(frontend_dir).parent)).lstrip(
            "/"
        )



from typing import cast

from django import template
from django.conf import settings
from django.utils.safestring import mark_safe

from frontend_kit.manifest import AssetTag
from frontend_kit.page import Page


class PageContextNotFoundError(Exception):
    def __init__(self, message: str = "Page not found in context") -> None:
        super().__init__(message)


register = template.Library()


def _get_csp_nonce(context: template.Context) -> str | None:
    if not getattr(settings, "DJFK_DEV_ENV", True):
        return None
    request = context.get("request")
    if not request:
        return None
    return getattr(request, "_csp_nonce", None)


@register.simple_tag(takes_context=True)
def fk_stylesheets(context: template.Context) -> str:
    if "page" not in context:
        raise PageContextNotFoundError

    page = cast(Page, context["page"])
    nonce = _get_csp_nonce(context)
    return mark_safe(
        "\n".join([tag.render(nonce=nonce) for tag in page.stylesheets])
    )


@register.simple_tag(takes_context=True)
def fk_preloads(context: template.Context) -> str:
    if "page" not in context:
        raise PageContextNotFoundError

    page = cast(Page, context["page"])
    return mark_safe("\n".join([tag.render() for tag in page.preload_imports]))


@register.simple_tag(takes_context=True)
def fk_head_scripts(context: template.Context) -> str:
    if "page" not in context:
        raise PageContextNotFoundError

    page = cast(Page, context["page"])
    nonce = _get_csp_nonce(context)
    return mark_safe(
        "\n".join([tag.render(nonce=nonce) for tag in page.head_imports])
    )


@register.simple_tag(takes_context=True)
def fk_body_scripts(context: template.Context) -> str:
    if "page" not in context:
        raise PageContextNotFoundError

    page = cast(Page, context["page"])
    nonce = _get_csp_nonce(context)
    return mark_safe(
        "\n".join([tag.render(nonce=nonce) for tag in page.body_imports])
    )


@register.simple_tag(takes_context=True)
def fk_custom_entry(context: template.Context, name: str) -> str:
    if "page" not in context:
        raise PageContextNotFoundError

    page = cast(Page, context["page"])
    nonce = _get_csp_nonce(context)
    files = [name + ".entry.js", name + ".entry.ts"]
    asset_tags: list[AssetTag] = []
    for file in files:
        try:
            tags = page.resolve_import(file_name=file)
        except FileNotFoundError:
            continue
        asset_tags.extend(list(tags))

    if not asset_tags:
        raise FileNotFoundError(
            f"Could not find entry file '{name}'. "
            f"Looked for files: {', '.join(files)}"
        )

    return mark_safe(
        "\n".join([tag.render(nonce=nonce) for tag in asset_tags])
    )



from pathlib import Path

from django.conf import settings


def get_frontend_dir_from_settings() -> Path:
    frontend_dir: str = getattr(settings, "DJFK_FRONTEND_DIR", "")
    if not frontend_dir:
        raise RuntimeError(
            "DJFK_FRONTEND_DIR is not set in settings.py, please set it to the "
            "output directory of your Vite build"
        )
    frontend_dir_path = Path(frontend_dir)
    if not frontend_dir_path.exists():
        raise RuntimeError(
            f"{frontend_dir} does not exist, "
            "please check DJFK_FRONTEND_DIR settings to ensure you "
            "pass correct dir"
        )

    if not frontend_dir_path.is_dir():
        raise RuntimeError(f"{frontend_dir} is not an directory")
    return frontend_dir_path
